#!/usr/bin/env python

# Modules to import
import sys

from libiSuperpose.config import *
from libiSuperpose.io import *
from libiSuperpose.core import *

# -- Globals --

# -- I/O functions --

# ARGUMENTS

def printUsage(exitCode):
    
    print >> sys.stderr, "Usage: superposePDBs.py [OPTIONS] -r [PDB file] -m [PDB file]"
    
    print >> sys.stderr, ""
    print >> sys.stderr, " -r : reference PDB file"
    print >> sys.stderr, " -m : mobile PDB file"
    
    print >> sys.stderr, ""
    print >> sys.stderr, "OPTIONS:"
    print >> sys.stderr, " -o : output fitted PDB file (stdout)"
    print >> sys.stderr, " -A : consider all continus residues (reference and mobile must have the same length)"
    print >> sys.stderr, " -f : input alignment (fasta format)"
    print >> sys.stderr, "      UPPER CASE, consider this residu in superposition"
    print >> sys.stderr, "      lower case, ignore residu in superposition"    
    print >> sys.stderr, " -g : output alignment file (stdout)"
    print >> sys.stderr, " -M : output TM-Score matrix (stdout)"
    print >> sys.stderr, " -R : output RMSd matrix (stdout)"
    print >> sys.stderr, " -G : output GDT_TS matrix (stdout)"
    print >> sys.stderr, " -s : reference rank for models or multiple PDB case (default is 0)"
    print >> sys.stderr, " -p : work directory"
    print >> sys.stderr, " -a : atoms to use for fitting process"
    print >> sys.stderr, "      Default is CA. Syntax is: -a N,CA,C,O"
    print >> sys.stderr, " -z : specify zones to fit"
    print >> sys.stderr, "      \"ref1from:ref1to::mobi1from:mobi1to|ref2from:ref2to::mobi2from:mobi2to\""
    print >> sys.stderr, "      No spaces are allowed. Indexes started at 0."
    print >> sys.stderr, " -Z : if zone(s) is precized output the full structure and not fitted zone(s) only"
    print >> sys.stderr, " -H : skip hetero residues"
    print >> sys.stderr, " -C : output ony CA atoms"
    print >> sys.stderr, " -h : prints help"
    print >> sys.stderr, " -v : verbose mode"
    
    sys.exit(exitCode)
    
    return 1

def argsParse( argList, verbose = 0):

    if verbose:
        print >> sys.stdout, "argsParse: Will parse arguments ..."

    # the arguments dictionnary
    argsDict = {}

    # -------- HELP
    try:
        argList.index("-h")
        help = 1
    except:
        help = 0

    if help:
        printUsage(1)
        
    # -------- VERBOSE
    try:
        argList.index("-v")
        argsDict['verbose'] = 1
    except:
        argsDict['verbose'] = 0

    # -------- REFERENCE PDB FILE
    try:
        refFNR = argList.index("-r")
        argsDict['refFN'] = argList[refFNR +1]
    except:
        print >> sys.stderr, "[ERROR] No reference pdb file precized !"
        printUsage(0)

    # -------- MOBILE PDB FILE
    try:
        mobiFNR = argList.index("-m")
        argsDict['mobiFN'] = argList[mobiFNR +1]
    except:
        # You can ommit mobile PDB files in case of
        # models in refererence file
        try:
            tmpPDB = loadPDBfile( argsDict['refFN'] )
        except:
            print >> sys.stderr, "[ERROR] Invalid reference pdb file format !"

        if not tmpPDB.nModels():                
            print >> sys.stderr, "[ERROR] No mobile pdb file precized !"
            printUsage(0)
            
        del(tmpPDB)
        argsDict['mobiFN'] = None
    # -------- OUTPUT TM Matrix
    try:
        try:
            outMATFNR = argList.index("-M")
            argsDict['outMAT'] = 1
            if outMATFNR +1 < len(argList) and argList[outMATFNR +1][0] != "-":
                argsDict['outMATFN'] = argList[outMATFNR +1]
            else:
                print >> sys.stderr, "No output file precized for TM Matrix, will use stdout ..."
                argsDict['outMATFN'] = None
        except:
            argsDict['outMAT'] = 0           
    except:
        if argsDict['outMAT']:
            print >> sys.stderr, "No output file precized for TM Matrix, will use stdout ..."
            argsDict['outMATFN'] = None
        else:
            argsDict['outMAT'] = 0
            
    # -------- OUTPUT RMSd Matrix
    try:
        try:
            outRMSdFNR = argList.index("-R")
            argsDict['outRMSd'] = 1
            if outRMSdFNR +1 < len(argList) and argList[outRMSdFNR +1][0] != "-":
                argsDict['outRMSdFN'] = argList[outRMSdFNR +1]
            else:
                print >> sys.stderr, "No output file precized for RMSd Matrix, will use stdout ..."
                argsDict['outRMSdFN'] = ""
        except:
            argsDict['outRMSd'] = 0           
    except:
        if argsDict['outRMSd']:
            print >> sys.stderr, "No output file precized for RMSd Matrix, will use stdout ..."
            argsDict['outRMSdFN'] = ""
        else:
            argsDict['outRMSd'] = 0

    # -------- OUTPUT GDT_TS Matrix
    try:
        try:
            outGDT_TSFNR = argList.index("-G")
            argsDict['outGDT_TS'] = 1
            if outGDT_TSFNR +1 < len(argList) and argList[outGDT_TSFNR +1][0] != "-":
                argsDict['outGDT_TSFN'] = argList[outGDT_TSFNR +1]
            else:
                print >> sys.stderr, "No output file precized for GDT_TS Matrix, will use stdout ..."
                argsDict['outGDT_TSFN'] = ""
        except:
            argsDict['outGDT_TS'] = 0           
    except:
        if argsDict['outGDT_TS']:
            print >> sys.stderr, "No output file precized for GDT_TS Matrix, will use stdout ..."
            argsDict['outGDT_TSFN'] = ""
        else:
            argsDict['outGDT_TS'] = 0
            
    # -------- OUTPUT PDB file
    try:
        outPDBFNR = argList.index("-o")
        argsDict['outPDBFN'] = argList[outPDBFNR +1]
    except:
        print >> sys.stderr, "No output file precized for PDB, will use stdout ..."
        argsDict['outPDBFN'] = None

    # -------- INPUT alignement file
    try:
        inALNFNR = argList.index("-f")
        argsDict['inALNFN'] = argList[inALNFNR +1]
    except:
        argsDict['inALNFN'] = None

    # -------- Linear superposition
    try:
        a = argList.index("-A")
        argsDict['linear'] = True
        print >> sys.stderr, "Linear superposition mode activated."
    except:
        argsDict['linear'] = False

    # -------- OUTPUT alignement file
    try:
        outALNFNR = argList.index("-g")
        argsDict['outALNFN'] = argList[outALNFNR +1]
    except:
        print >> sys.stderr, "No output file precized for alignment, will use stdout ..."
        argsDict['outALNFN'] = None

    # -------- Rank of the reference (Models or multiple PDB case)
    try:
        refRkR = argList.index("-s")
        argsDict['refRk'] = int(argList[refRkR +1])
    except:
        argsDict['refRk'] = 0

    # -------- working directory
    try:
        wPath = argList.index("-p")
        argsDict['wPath'] = argList[wPath +1]
    except:
        argsDict['wPath'] = "./"

    # -------- atoms to use
    try:
        atmsRk = argList.index("-a")
        # parse proposed atoms
        atms = []
        for aAtm in argList[atmsRk +1].split(","):
            if aAtm not in AVAILATMS:
                print >> sys.stderr, "Invalid atom name '%s' will be ignored !"%aAtm
                continue
            if aAtm != "" and aAtm != " ":
                atms.append(aAtm)
        
        if len(atms):
            print >> sys.stderr, "Input atoms used for superposition: ",atms
            argsDict['atms'] = atms
        else:
            print >> sys.stderr, "No valid atoms precized for superposition, will use CA."
            argsDict['atms'] = ['CA']
    except:
        print >> sys.stderr, "No valid atoms precized for superposition, will use CA."
        argsDict['atms'] = ['CA']
    
    # -------- restrict superposition to zones
    try:
        znDef = argList.index("-z")
        argsDict['zones'] = parseUsrDefZn( argList[ znDef  +1], argsDict['verbose'] )
    except Exception:
        argsDict['zones'] = {}
        argsDict['zones']['ref']  = [[0,-1]]
        argsDict['zones']['mobi'] = [[0,-1]]

    try:
        znDef = argList.index("-Z")
        argsDict['rawTrans'] = True
    except Exception:
        argsDict['rawTrans'] = False
        
    # -------- Skip hetero atoms
    try:
        znDef = argList.index("-H")
        argsDict['hetSkip'] = 1
    except:
        argsDict['hetSkip'] = 0
        
    # -------- Skip hetero atoms
    try:
        znDef = argList.index("-C")
        argsDict['CAonly'] = 1
    except:
        argsDict['CAonly'] = 0

    return argsDict

def zonesOverlap( theZones ):
    """
    We test if zones overlaps.
    """

    for aZone1 in range( 0, len(theZones)-1, 1 ):
        for aZone2 in range( aZone1+1, len(theZones), 1 ):
            if ( theZones[ aZone1 ][0] >= theZones[ aZone2 ][0] and theZones[ aZone1 ][0] <= theZones[ aZone2 ][1] ) or ( theZones[ aZone1 ][1] >= theZones[ aZone2 ][0] and theZones[ aZone1 ][1] <= theZones[ aZone2 ][1] ):
                return True

    return False


def parseUsrDefZn( znDef, verbose = 1 ):
    """
    Parse user defined zones.
    
    syntax is : 
         zone1|zone2|...|zoneN
    with zoneX    = RefZoneX::MobiZoneX
         RefZoneX = from-to
    exemple:
         12:18::13:19|24:30::25:31
    """
    if verbose:
        print >> sys.stderr, "parseUsrDefZn: will parse user zone definition: %s " % znDef

    zones = []
    a = znDef.split("|")
    for b in a:
        c = b.split("::")
        tmp1 = []
        for d in c:
            e = d.split(":")
            tmp2 = []
            for f in e:
                try:
                    tmp2.append( int(f) )
                except:
                    print >> sys.stderr, "Error: Invalid pattern for zones !"
                    sys.exit(0)
            tmp1.append(tmp2)
        zones.append(tmp1)

    # Correctly represents selections
    # Ref  = index 0
    # Mobi = index 1
    sZones = {}
    sZones['ref']  = []
    sZones['mobi'] = []

    for aZone in zones:
        sZones['ref'].append( aZone[0] )
        if len(aZone) > 1:
            sZones['mobi'].append( aZone[1] )
        else:
            print >> sys.stderr, "Error: %d zone(s) defined for mobile and %d for reference structure !" % (len(sZones['mobi']), len(sZones['ref']))
            sys.exit(0)

        # Check zones length
        if ( aZone[0][1] - aZone[0][0] ) != ( aZone[1][1] - aZone[1][0] ):
            print >> sys.stderr, "Error: zones lengths differs: [%d-%d] vs [%d-%d] !" % ( aZone[0][0], aZone[0][1], aZone[1][0], aZone[1][1] )
            sys.exit(0)

    if verbose:
        print >> sys.stderr, " -> ", sZones

    # check lists overlapping
    if zonesOverlap( sZones['ref'] ):
        print >> sys.stderr, "Error: zones definition must not overlap ! Reference:", sZones['ref']
        sys.exit(0)
    if zonesOverlap( sZones['mobi'] ):
        print >> sys.stderr, "Error: zones definition must not overlap ! Mobile:", sZones['mobi']
        sys.exit(0)
        
    return sZones
    

# -- main purpose --
def process( argsDict ):

    verbose = argsDict['verbose']

    # atoms used to superpose
    # atms = ["CA"]
    atms = argsDict['atms']

    # files list to delete
    f2Del = []

    # files list to concatenate
    f2Cat = []

    # TM-Scores Matrix
    TMmat = {}
    TMmat['scores'] = []
    TMmat['references'] = []
    TMmat['mobiles'] = []
        
    # CA RMSd Matrix
    CARMSdMat_zone = []
    CARMSdMat = []
    CAGDT_TSMat = []

    refRk = argsDict['refRk']
    
    # Parse the input alignement if any
    if argsDict['inALNFN']:
        seqs = parseFastaAlign( argsDict['inALNFN'], verbose = verbose )
        
    # Launch reference PDB files
    PDB_Ref_Files = extractPDB( argsDict['refFN'], zones = argsDict['zones']['ref'], hetSkip = argsDict['hetSkip'], verbose = verbose )
    
    # Launch mobile PDB files if any
    if argsDict['mobiFN']:
        PDB_Mobi_Files = extractPDB( argsDict['mobiFN'], zones = argsDict['zones']['mobi'], hetSkip = argsDict['hetSkip'], verbose = verbose )
        if argsDict['rawTrans']:
            PDB_Mobi_Files_Raw = extractPDB( argsDict['mobiFN'], hetSkip = argsDict['hetSkip'], verbose = verbose )
        else:
            PDB_Mobi_Files_Raw = [None] * len(PDB_Mobi_Files)
        meth = PDB_Mobi_Files[0]
    else:
        # User want to fit a particular PDB item on the others        
        meth = PDB_Ref_Files[0]
        
        # NB: first rank is the method
        PDB_Mobi_Files = PDB_Ref_Files[:refRk+1] + PDB_Ref_Files[refRk+2:]
        if argsDict['rawTrans']:
            tmpExtract = extractPDB( argsDict['refFN'], hetSkip = argsDict['hetSkip'], verbose = verbose )
            PDB_Mobi_Files_Raw = tmpExtract[:refRk+1] + tmpExtract[refRk+2:]
        else:
            PDB_Mobi_Files_Raw = [None] * len(PDB_Mobi_Files)
        tmp = PDB_Ref_Files[refRk+1]
        PDB_Ref_Files = [1,tmp]
    
    # if an alignment is precized rank must be the same
    if argsDict['inALNFN']:
        mobiSeqs = seqs[:refRk] + seqs[refRk+1:]
        refSeq = seqs[refRk]
        aSeqMaskRef  = fasta2SeqMask( refSeq[1] )

    # Does the mobile file contains models
    if meth == 3:
        isModel = 1
    else:
        isModel = 0

    TMmat['references'] += list(PDB_Ref_Files[1:])
    TMmat['mobiles'] +=  list(PDB_Mobi_Files[1:])

    aRefRk  = 0
    refIds  = []
    mobiIds = []
    isSetMobiIds = 0
    
    # For each Reference/Mobile combinaison, process:
    # NOTE:
    # Some problems will occur when multiple PDB or models
    # are given in reference and mobile and an alignement is
    # precized.
    for aRefFN in list(PDB_Ref_Files[1:]):
        aMobiRk = 0
        aTMmat = []
        aCARMSdMat_zone = []
        aCARMSdMat = []
        aCAGDT_TSMat = []

        # One output file by reference, so
        f2Cat = []

        # load PDB instances 
        aPDB_Ref  = loadPDBfile( aRefFN, hetSkip = argsDict['hetSkip'], verbose = verbose )

        # get the reference id
        # if not in header, take the filename
        if aPDB_Ref.id:
            aRefId = aPDB_Ref.id[:4]
        else:
            aRefId = aRefFN.split("/")[-1][:4]
            
        aRefId += "_" + str( aRefRk )
        
        refIds.append( aRefId )

        if argsDict['outPDBFN']:
            # We have multiple references
            if len(list(PDB_Ref_Files[1:])) >1:
                if argsDict['outPDBFN'][-4:] == ".pdb":
                    aRefOutSuppFN = argsDict['outPDBFN'][:-4] + "_ref%s.pdb"%aRefId
                else:
                    aRefOutSuppFN = argsDict['outPDBFN'] + "_ref%s.pdb"%aRefId
            else:
                aRefOutSuppFN = argsDict['outPDBFN']
        else:
            aRefOutSuppFN = "superposed_ref%s.pdb"%aRefId
        
        for aMobi in zip( list(PDB_Mobi_Files[1:]), list(PDB_Mobi_Files_Raw[1:]) ):

            aMobiFN = aMobi[0]
            aMobiFNRaw = aMobi[1]

            if verbose:
                print >> sys.stderr, "\n>> Ref: %s Mobi: %s\n"%(aRefFN, aMobiFN)

            # load PDB instances 
            aPDB_Mobi = loadPDBfile( aMobiFN, hetSkip = argsDict['hetSkip'], verbose = verbose)
            if aMobiFNRaw:
                aPDB_Mobi_Raw = loadPDBfile( aMobiFNRaw, hetSkip = argsDict['hetSkip'], verbose = verbose)
            else:
                aPDB_Mobi_Raw = None

            if verbose:
                print >> sys.stderr, "Ref id: %s Mobi id: %s"%(aPDB_Ref.id, aPDB_Mobi.id)
                
            if not isSetMobiIds:
                if aPDB_Mobi.id:
                    aMobiId = aPDB_Mobi.id[:4]
                else:
                    aMobiId = aMobiFN.split("/")[-1][:4]
                aMobiId += "_"+ str( aMobiRk )
                mobiIds.append( aMobiId )
            else:
                aMobiId = mobiIds[aMobiRk]

            # By default, we initialize our TMscore object
            aTM = {'TMScore':0.0}

            # Linear superposition
            # We consider all residues
            if argsDict['linear']:
                if len(aaseq2(aPDB_Ref, atms)) == len(aaseq2(aPDB_Mobi,atms)) :
                    aSeqMaskRef = aSeqMaskMobi = [1] * len(aaseq2(aPDB_Mobi,atms))
                else:
                    print >> sys.stderr, "Error: for linear superposition structures must have the same length !"
                    print >> sys.stderr, "       Reference: %d" % len(aaseq2(aPDB_Ref,atms))
                    print >> sys.stderr, "       Mobile:    %d" % len(aaseq2(aPDB_Mobi,atms))
                    sys.exit(0)
            # is there an alignment proposed ?
            elif argsDict['inALNFN'] == None:
                # if no alignment, use TMalign
                aTM = twoPDBTMalign2( aRefFN, aMobiFN, wPath = argsDict['wPath'], verbose = verbose )
                (aSeqMaskRef, aSeqMaskMobi) = TM2seqMasks( aTM, aaseq2(aPDB_Ref,atms), aaseq2(aPDB_Mobi,atms), verbose = verbose )
            else:
                # use input alignment
                #aSeqMaskRef  = fasta2SeqMask2( refSeq[1], mobiSeqs[aMobiRk][1] )
                aSeqMaskRef  = fasta2SeqMask( refSeq[1] )
                aSeqMaskMobi = fasta2SeqMask( mobiSeqs[aMobiRk][1] )
                
            if argsDict['outALNFN']:
                TMaln2fastaOut( aTM, aRefFN, aMobiFN, argsDict['outALNFN'], ids = [aRefId,aMobiId])
            aTMmat.append(aTM['TMScore'])
            
            # Then superpose with QBestFit
            aPDB_MobiTrans, aCARSMd_zone = twoPDBspp(aPDB_Ref, aPDB_Mobi, aSeqMaskRef, aSeqMaskMobi, atms, wPath = argsDict['wPath'], PDB_TM = aPDB_Mobi_Raw, verbose = verbose)
            aCARMSdMat_zone.append(aCARSMd_zone)

            # If the 2 files lenghts are identical: compute RMSd
            if len(aPDB_Ref) == len(aPDB_Mobi):
                if verbose:
                    print >> sys.stderr, "PDB files have the same length. Will compute cRMSd and GDT_TS."
                #aRMSd, aGDT_TS = twoPDBCompare(aPDB_Ref, aPDB_MobiTrans, atms, wPath = argsDict['wPath'], verbose = verbose)
                aRMSd, aGDT_TS = twoPDBCompare(aPDB_Ref, aPDB_Mobi, atms, wPath = argsDict['wPath'], verbose = verbose)
                aCARMSdMat.append( aRMSd )
                aCAGDT_TSMat.append( aGDT_TS )
            else:
                if verbose:
                    print >> sys.stderr, "PDB files lengths are different. Cannot compute cRMSd and GDT_TS."
                aCARMSdMat.append( -1.0 )
                aCAGDT_TSMat.append( -1.0 )

            # ouput superposed mobile PDB file
            if aMobiFN[-4:] == ".pdb":
                aOutFN = aMobiFN[:-4] + "_supp.pdb"
            else:
                aOutFN = aMobiFN + "_supp.pdb"

            if meth == 1:
                aOutFN = argsDict['outPDBFN']

            # We need to force output
            outputPDB( aOutFN, aPDB_MobiTrans, CAonly=argsDict['CAonly'] )

            # if not simple PDB file method
            if meth != 1:
                
                # Don't forget the reference
                if aMobiRk == argsDict['refRk']:
                    f2Del.append( aRefFN )
                    #f2Cat.append( aRefFN )
                    
                f2Del.append( aMobiFN )
                f2Del.append( aMobiFNRaw )
                f2Del.append( aOutFN )              
                f2Cat.append( aOutFN )
                
            aMobiRk += 1
            
        isSetMobiIds += 1
        aRefRk += 1
        
        TMmat['scores'].append(aTMmat)
        CARMSdMat_zone.append( aCARMSdMat_zone )
        CARMSdMat.append( aCARMSdMat )
        CAGDT_TSMat.append( aCAGDT_TSMat )

        # Temporary files management for
        # other methods than simple PDB
        # if meth != 1 and not argsDict['outMAT']:
        isModel = 1
        if meth != 1:
            catSuperposedFiles( aRefOutSuppFN, f2Cat, isModel, mobiIds, hetSkip = argsDict['hetSkip'], CAonly=argsDict['CAonly'], verbose = verbose )
        
    # Temporary files management for
    # other methods than simple PDB
    # if meth != 1 and not argsDict['outMAT']:
    #if meth != 1 :
    #    catSuperposedFiles( argsDict['outPDBFN'], f2Cat, isModel, mobiIds)
        
    f2Del += PDB_Ref_Files[1:] + PDB_Mobi_Files[1:]
    eraseTempFiles( f2Del )
    
    if argsDict['outMAT']:
        TMmatOut( TMmat, ids = [refIds, mobiIds], outFN = argsDict['outMATFN'] )
        
    if argsDict['outRMSd']:
        hcomm = "# RMSd computed on %s\n# NB: fitted zones only\n" % str(atms)
        outMat(CARMSdMat_zone, ids = [refIds, mobiIds], hcomm = hcomm, outFN = argsDict['outRMSdFN'] )
        hcomm = "# RMSd computed on %s\n# NB: all residues (for identical PDB lengths)\n" % str(atms)
        outMat(CARMSdMat, ids = [refIds, mobiIds], hcomm = hcomm, outFN = argsDict['outRMSdFN'] )
    
    if argsDict['outGDT_TS']:
        hcomm = "# GDT_TS\n# NB: computed with iSuperposed structures\n" 
        outMat(CAGDT_TSMat, ids = [refIds, mobiIds], hcomm = hcomm, outFN = argsDict['outGDT_TSFN'] )

    # Simple one to one comparison
    # Get infos
    if meth == 1:
        print >> sys.stdout, "# %18s\t%20s %8s %8s %8s %8s" % ( "file1", "file2", "cRMSd_f", "TMScore", "cRMSd", "GDT_TS" )
        print >> sys.stdout, "%20s\t%20s %8.3f %8.3f %8.3f %8.3f" % ( 
            argsDict['refFN'], argsDict['mobiFN'], 
            CARMSdMat_zone[0][0], TMmat['scores'][0][0], CARMSdMat[0][0], CAGDT_TSMat[0][0]
            )

    return 1

def main():
    """
    Launch the main purpose: superpose 2 pdb files with(out)
    an imposed alignment. If no alignment precized, TMalign
    will be used. In all cases, the QBestFit procedure will
    superpose selected coordinates. Source code is available
    at:

    http://bioserv.rpbs.jussieu.fr/software/QBestFit.c    
    """

    # Parse arguments
    argsDict = argsParse( sys.argv )
        
    # process
    ret = process( argsDict )

    return ret

# GO !
if __name__ == "__main__":
    main()
